from __future__ import annotations

import functools
import pickle

import pytest

from dask._expr import Expr, ProhibitReuse, SingletonExpr, _ExprSequence
from dask._task_spec import DataNode


class MyExpr(Expr):
    _parameters = ["foo", "bar"]


class MyExpr2(MyExpr):
    # A subclass that inherits parameters
    pass


def test_setattr():
    e = MyExpr(foo=1, bar=2)
    e.bar = 3
    assert e.bar == 3
    with pytest.raises(AttributeError):
        e.baz = 4


def test_setattr2():
    e = MyExpr2(foo=1, bar=2)
    e.bar = 3
    assert e.bar == 3
    with pytest.raises(AttributeError):
        e.baz = 4


class MyExprCachedProperty(Expr):
    called_cached_property = False
    _parameters = ["foo", "bar"]

    @property
    def baz(self):
        return self.foo + self.bar

    @functools.cached_property
    def cached_property(self):
        if MyExprCachedProperty.called_cached_property:
            raise RuntimeError("No!")
        MyExprCachedProperty.called_cached_property = True
        return self.foo + self.bar


@pytest.mark.slow()
def test_pickle_cached_properties():
    pytest.importorskip("distributed")
    from distributed import Nanny
    from distributed.utils_test import gen_cluster

    @gen_cluster(client=True, Worker=Nanny, nthreads=[("", 1)])
    async def test(c, s, a):

        expr = MyExprCachedProperty(foo=1, bar=2)
        for _ in range(10):
            assert expr.baz == 3
            assert expr.cached_property == 3

        assert MyExprCachedProperty.called_cached_property is True

        rt = pickle.loads(pickle.dumps(expr))
        assert rt.cached_property == 3
        assert MyExprCachedProperty.called_cached_property is True

        # But this does
        expr3 = MyExprCachedProperty(foo=1, bar=3)
        with pytest.raises(RuntimeError):
            expr3.cached_property

        def f(expr):
            # We want the cache to be part of the pickle, i.e. this is a
            # different process such that the type is reset and the property can
            # be accessed without side effects
            assert MyExprCachedProperty.called_cached_property is False
            assert expr.cached_property == 3
            assert MyExprCachedProperty.called_cached_property is False

        await c.submit(f, expr)

    test()


class MySingleton(SingletonExpr): ...


class MySingletonWithCustomInit(SingletonExpr):
    def __init__(self, *args, **kwargs): ...


class MySingletonInheritsCustomInit(MySingletonWithCustomInit): ...


class Mixin:
    def __init__(self, *args, **kwargs): ...


class MySingletonInheritsCustomInitAsMixin(MySingleton, Mixin): ...


def test_singleton_expr():
    assert MySingleton(1, 2) is MySingleton(1, 2)
    # We don't want to deduplicate if there is an __init__ since that may
    # mutatate our singleton reference and we have no way to know
    assert MySingletonWithCustomInit(1, 2) is not MySingletonWithCustomInit(1, 2)
    assert MySingletonInheritsCustomInit(1, 2) is not MySingletonInheritsCustomInit(
        1, 2
    )
    assert MySingletonInheritsCustomInitAsMixin(
        1, 2
    ) is not MySingletonInheritsCustomInitAsMixin(1, 2)


@pytest.mark.slow()
def test_refcounting_futures():
    pd = pytest.importorskip("pandas")
    dd = pytest.importorskip("dask.dataframe")
    distributed = pytest.importorskip("distributed")

    # See https://github.com/dask/distributed/issues/9041
    # Didn't reproduce with any of our fixtures
    with distributed.Client(
        n_workers=2, worker_class=distributed.Worker, dashboard_address=":0"
    ) as client:

        def gen(i):
            return pd.DataFrame({"A": [i]}, index=[i])

        futures = [client.submit(gen, i) for i in range(3)]

        meta = gen(0)[:0]
        df = dd.from_delayed(futures, meta)
        df.compute()

        del futures

        df.compute()


class FooExpr(Expr):
    def _layer(self) -> dict:
        return {"foo": DataNode("foo", 42)}


def test_prohibit_reuse():
    once = FooExpr()
    ProhibitReuse._ALLOWED_TYPES.append(FooExpr)
    try:
        dsk = _ExprSequence(once, ProhibitReuse(once)).optimize().__dask_graph__()

        assert len(dsk) == 2
        first = dsk.pop("foo")()
        key, val = dsk.popitem()
        assert key.startswith("foo") and key != "foo"
        # We don't want to chain anything but actually _hide_ the task
        assert not val.dependencies
        # Task is wrapped
        assert val() is first
    finally:
        ProhibitReuse._ALLOWED_TYPES.remove(FooExpr)



from __future__ import annotations

import pytest

from dask.dataframe.dask_expr import from_pandas
from dask.dataframe.dask_expr._categorical import GetCategories
from dask.dataframe.dask_expr.tests._util import _backend_library, assert_eq

# Set DataFrame backend for this module
pd = _backend_library()


@pytest.fixture
def pdf():
    pdf = pd.DataFrame({"x": [1, 2, 3, 4, 1, 2], "y": "bcbbbc"})
    return pdf


@pytest.fixture
def df(pdf):
    yield from_pandas(pdf, npartitions=2)


def test_set_categories(pdf):
    pdf = pdf.astype("category")
    df = from_pandas(pdf, npartitions=2)
    assert df.x.cat.known
    assert_eq(df.x.cat.codes, pdf.x.cat.codes)
    ser = df.x.cat.as_unknown()
    assert not ser.cat.known
    ser = ser.cat.as_known()
    assert_eq(ser.cat.categories, pd.Index([1, 2, 3, 4]))
    ser = ser.cat.set_categories([1, 2, 3, 5, 4])
    assert_eq(ser.cat.categories, pd.Index([1, 2, 3, 5, 4]))
    assert not ser.cat.ordered


def test_categorize(df, pdf):
    df = df.categorize()

    assert df.y.cat.known
    assert_eq(df, pdf.astype({"y": "category"}), check_categorical=False)


def test_get_categories_simplify_adds_projection(df):
    optimized = GetCategories(
        df, columns=["y"], index=False, split_every=None
    ).simplify()
    expected = GetCategories(
        df[["y"]].simplify(), columns=["y"], index=False, split_every=None
    )
    assert optimized._name == expected._name


def test_categorical_set_index():
    df = pd.DataFrame({"x": [1, 2, 3, 4], "y": ["a", "b", "b", "c"]})
    df["y"] = pd.Categorical(df["y"], categories=["a", "b", "c"], ordered=True)
    a = from_pandas(df, npartitions=2)

    b = a.set_index("y", divisions=["a", "b", "c"], npartitions=a.npartitions)
    d1, d2 = b.get_partition(0), b.get_partition(1)
    assert list(d1.index.compute(fuse=False)) == ["a"]
    assert sorted(d2.index.compute()) == ["b", "b", "c"]


def test_categorize_drops_category_columns():
    pdf = pd.DataFrame({"a": [1, 2, 1, 2, 3], "b": 1})
    df = from_pandas(pdf)
    df = df.categorize(columns=["a"])
    result = df["b"].to_frame()
    expected = pdf["b"].to_frame()
    assert_eq(result, expected)



from __future__ import annotations

import contextlib
import importlib
import numbers
from itertools import product
from numbers import Integral
from threading import Lock

import numpy as np

from dask._collections import new_collection
from dask.array._array_expr._collection import Array
from dask.array._array_expr._io import IO
from dask.array.backends import array_creation_dispatch
from dask.array.core import asarray, broadcast_shapes, normalize_chunks
from dask.array.creation import arange
from dask.array.utils import asarray_safe
from dask.tokenize import tokenize
from dask.utils import cached_property, derived_from, random_state_data, typename


class Generator:
    """
    Container for the BitGenerators.

    ``Generator`` exposes a number of methods for generating random
    numbers drawn from a variety of probability distributions and serves
    as a replacement for ``RandomState``. The main difference between the
    two is that ``Generator`` relies on an additional ``BitGenerator`` to
    manage state and generate the random bits, which are then transformed
    into random values from useful distributions. The default ``BitGenerator``
    used by ``Generator`` is ``PCG64``. The ``BitGenerator`` can be changed
    by passing an instantiated ``BitGenerator`` to ``Generator``.

    The function :func:`dask.array.random.default_rng` is the recommended way
    to instantiate a ``Generator``.

    .. warning::

       No Compatibility Guarantee.

       ``Generator`` does not provide a version compatibility guarantee. In
       particular, as better algorithms evolve the bit stream may change.

    Parameters
    ----------
    bit_generator : BitGenerator
        BitGenerator to use as the core generator.

    Notes
    -----
    In addition to the distribution-specific arguments, each ``Generator``
    method takes a keyword argument `size` that defaults to ``None``. If
    `size` is ``None``, then a single value is generated and returned. If
    `size` is an integer, then a 1-D array filled with generated values is
    returned. If `size` is a tuple, then an array with that shape is
    filled and returned.

    The Python stdlib module `random` contains pseudo-random number generator
    with a number of methods that are similar to the ones available in
    ``Generator``. It uses Mersenne Twister, and this bit generator can
    be accessed using ``MT19937``. ``Generator``, besides being
    Dask-aware, has the advantage that it provides a much larger number
    of probability distributions to choose from.

    All ``Generator`` methods are identical to ``np.random.Generator`` except
    that they also take a `chunks=` keyword argument.

    ``Generator`` does not guarantee parity in the generated numbers
    with any third party library. In particular, numbers generated by
    `Dask` and `NumPy` will differ even if they use the same seed.

    Examples
    --------
    >>> from numpy.random import PCG64
    >>> from dask.array.random import Generator
    >>> rng = Generator(PCG64())
    >>> rng.standard_normal().compute() # doctest: +SKIP
    array(0.44595957)  # random

    See Also
    --------
    default_rng : Recommended constructor for `Generator`.
    np.random.Generator
    """

    def __init__(self, bit_generator):
        self._bit_generator = bit_generator

    def __str__(self):
        _str = self.__class__.__name__
        _str += "(" + self._bit_generator.__class__.__name__ + ")"
        return _str

    @property
    def _backend_name(self):
        # Assumes typename(self._RandomState) starts with an
        # array-library name (e.g. "numpy" or "cupy")
        return typename(self._bit_generator).split(".")[0]

    @property
    def _backend(self):
        # Assumes `self._backend_name` is an importable
        # array-library name (e.g. "numpy" or "cupy")
        return importlib.import_module(self._backend_name)

    @derived_from(np.random.Generator, skipblocks=1)
    def beta(self, a, b, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "beta", a, b, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def binomial(self, n, p, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "binomial", n, p, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def chisquare(self, df, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "chisquare", df, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def choice(
        self,
        a,
        size=None,
        replace=True,
        p=None,
        axis=0,
        shuffle=True,
        chunks="auto",
    ):
        (
            a,
            size,
            replace,
            p,
            axis,
            chunks,
            meta,
            dependencies,
        ) = _choice_validate_params(self, a, size, replace, p, axis, chunks)

        return new_collection(
            RandomChoiceGenerator(
                a.expr, chunks, meta, self._bit_generator, replace, p, axis, shuffle
            )
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def exponential(self, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "exponential", scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def f(self, dfnum, dfden, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "f", dfnum, dfden, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def gamma(self, shape, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "gamma", shape, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def geometric(self, p, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "geometric", p, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def gumbel(self, loc=0.0, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "gumbel", loc, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def hypergeometric(self, ngood, nbad, nsample, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self,
            "hypergeometric",
            ngood,
            nbad,
            nsample,
            size=size,
            chunks=chunks,
            **kwargs,
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def integers(
        self,
        low,
        high=None,
        size=None,
        dtype=np.int64,
        endpoint=False,
        chunks="auto",
        **kwargs,
    ):
        return _wrap_func(
            self,
            "integers",
            low,
            high=high,
            size=size,
            dtype=dtype,
            endpoint=endpoint,
            chunks=chunks,
            **kwargs,
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def laplace(self, loc=0.0, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "laplace", loc, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def logistic(self, loc=0.0, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "logistic", loc, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def lognormal(self, mean=0.0, sigma=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "lognormal", mean, sigma, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def logseries(self, p, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "logseries", p, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def multinomial(self, n, pvals, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self,
            "multinomial",
            n,
            pvals,
            size=size,
            chunks=chunks,
            extra_chunks=((len(pvals),),),
            **kwargs,
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def multivariate_hypergeometric(
        self, colors, nsample, size=None, method="marginals", chunks="auto", **kwargs
    ):
        return _wrap_func(
            self,
            "multivariate_hypergeometric",
            colors,
            nsample,
            size=size,
            method=method,
            chunks=chunks,
            **kwargs,
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def negative_binomial(self, n, p, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "negative_binomial", n, p, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def noncentral_chisquare(self, df, nonc, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "noncentral_chisquare", df, nonc, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def noncentral_f(self, dfnum, dfden, nonc, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "noncentral_f", dfnum, dfden, nonc, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def normal(self, loc=0.0, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "normal", loc, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def pareto(self, a, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "pareto", a, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def permutation(self, x):
        from dask.array.slicing import shuffle_slice

        if self._backend_name == "cupy":
            raise NotImplementedError(
                "`Generator.permutation` not supported for cupy-backed "
                "Generator objects. Use the 'numpy' array backend to "
                "call `dask.array.random.default_rng`, or pass in "
                " `numpy.random.PCG64()`."
            )

        if isinstance(x, numbers.Number):
            x = arange(x, chunks="auto")

        index = self._backend.arange(len(x))
        _shuffle(self._bit_generator, index)
        return shuffle_slice(x, index)

    @derived_from(np.random.Generator, skipblocks=1)
    def poisson(self, lam=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "poisson", lam, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def power(self, a, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "power", a, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def random(self, size=None, dtype=np.float64, out=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "random", size=size, dtype=dtype, out=out, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def rayleigh(self, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "rayleigh", scale, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def standard_cauchy(self, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "standard_cauchy", size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def standard_exponential(self, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "standard_exponential", size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def standard_gamma(self, shape, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "standard_gamma", shape, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def standard_normal(self, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "standard_normal", size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def standard_t(self, df, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "standard_t", df, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def triangular(self, left, mode, right, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "triangular", left, mode, right, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def uniform(self, low=0.0, high=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "uniform", low, high, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def vonmises(self, mu, kappa, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "vonmises", mu, kappa, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.Generator, skipblocks=1)
    def wald(self, mean, scale, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "wald", mean, scale, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def weibull(self, a, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "weibull", a, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.Generator, skipblocks=1)
    def zipf(self, a, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "zipf", a, size=size, chunks=chunks, **kwargs)


def default_rng(seed=None):
    """
    Construct a new Generator with the default BitGenerator (PCG64).

    Parameters
    ----------
    seed : {None, int, array_like[ints], SeedSequence, BitGenerator, Generator}, optional
        A seed to initialize the `BitGenerator`. If None, then fresh,
        unpredictable entropy will be pulled from the OS. If an ``int`` or
        ``array_like[ints]`` is passed, then it will be passed to
        `SeedSequence` to derive the initial `BitGenerator` state. One may
        also pass in a `SeedSequence` instance.
        Additionally, when passed a `BitGenerator`, it will be wrapped by
        `Generator`. If passed a `Generator`, it will be returned unaltered.

    Returns
    -------
    Generator
        The initialized generator object.

    Notes
    -----
    If ``seed`` is not a `BitGenerator` or a `Generator`, a new
    `BitGenerator` is instantiated. This function does not manage a default
    global instance.

    Examples
    --------
    ``default_rng`` is the recommended constructor for the random number
    class ``Generator``. Here are several ways we can construct a random
    number generator using ``default_rng`` and the ``Generator`` class.

    Here we use ``default_rng`` to generate a random float:

    >>> import dask.array as da
    >>> rng = da.random.default_rng(12345)
    >>> print(rng)
    Generator(PCG64)
    >>> rfloat = rng.random().compute()
    >>> rfloat
    array(0.86999885)
    >>> type(rfloat)
    <class 'numpy.ndarray'>

    Here we use ``default_rng`` to generate 3 random integers between 0
    (inclusive) and 10 (exclusive):

    >>> import dask.array as da
    >>> rng = da.random.default_rng(12345)
    >>> rints = rng.integers(low=0, high=10, size=3).compute()
    >>> rints
    array([2, 8, 7])
    >>> type(rints[0])
    <class 'numpy.int64'>

    Here we specify a seed so that we have reproducible results:

    >>> import dask.array as da
    >>> rng = da.random.default_rng(seed=42)
    >>> print(rng)
    Generator(PCG64)
    >>> arr1 = rng.random((3, 3)).compute()
    >>> arr1
    array([[0.91674416, 0.91098667, 0.8765925 ],
           [0.30931841, 0.95465607, 0.17509458],
           [0.99662814, 0.75203348, 0.15038118]])

    If we exit and restart our Python interpreter, we'll see that we
    generate the same random numbers again:

    >>> import dask.array as da
    >>> rng = da.random.default_rng(seed=42)
    >>> arr2 = rng.random((3, 3)).compute()
    >>> arr2
    array([[0.91674416, 0.91098667, 0.8765925 ],
           [0.30931841, 0.95465607, 0.17509458],
           [0.99662814, 0.75203348, 0.15038118]])

    See Also
    --------
    np.random.default_rng
    """
    if hasattr(seed, "capsule"):
        # We are passed a BitGenerator, so just wrap it
        return Generator(seed)
    elif isinstance(seed, Generator):
        # Pass through a Generator
        return seed
    elif hasattr(seed, "bit_generator"):
        # a Generator. Just not ours
        return Generator(seed.bit_generator)
    # Otherwise, use the backend-default BitGenerator
    return Generator(array_creation_dispatch.default_bit_generator(seed))


class RandomState:
    """
    Mersenne Twister pseudo-random number generator

    This object contains state to deterministically generate pseudo-random
    numbers from a variety of probability distributions.  It is identical to
    ``np.random.RandomState`` except that all functions also take a ``chunks=``
    keyword argument.

    Parameters
    ----------
    seed: Number
        Object to pass to RandomState to serve as deterministic seed
    RandomState: Callable[seed] -> RandomState
        A callable that, when provided with a ``seed`` keyword provides an
        object that operates identically to ``np.random.RandomState`` (the
        default).  This might also be a function that returns a
        ``mkl_random``, or ``cupy.random.RandomState`` object.

    Examples
    --------
    >>> import dask.array as da
    >>> state = da.random.RandomState(1234)  # a seed
    >>> x = state.normal(10, 0.1, size=3, chunks=(2,))
    >>> x.compute()
    array([10.01867852, 10.04812289,  9.89649746])

    See Also
    --------
    np.random.RandomState
    """

    def __init__(self, seed=None, RandomState=None):
        self._numpy_state = np.random.RandomState(seed)
        self._RandomState = (
            array_creation_dispatch.RandomState if RandomState is None else RandomState
        )

    @property
    def _backend(self):
        # Assumes typename(self._RandomState) starts with
        # an importable array-library name (e.g. "numpy" or "cupy")
        _backend_name = typename(self._RandomState).split(".")[0]
        return importlib.import_module(_backend_name)

    def seed(self, seed=None):
        self._numpy_state.seed(seed)

    @derived_from(np.random.RandomState, skipblocks=1)
    def beta(self, a, b, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "beta", a, b, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def binomial(self, n, p, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "binomial", n, p, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def chisquare(self, df, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "chisquare", df, size=size, chunks=chunks, **kwargs)

    with contextlib.suppress(AttributeError):

        @derived_from(np.random.RandomState, skipblocks=1)
        def choice(self, a, size=None, replace=True, p=None, chunks="auto"):
            # TODO(expr): Fix this, this is ugly and should happen later
            (
                a,
                size,
                replace,
                p,
                axis,  # np.random.RandomState.choice does not use axis
                chunks,
                meta,
                dependencies,
            ) = _choice_validate_params(self, a, size, replace, p, 0, chunks)

            return new_collection(
                RandomChoice(a.expr, chunks, meta, self._numpy_state, replace, p)
            )

    @derived_from(np.random.RandomState, skipblocks=1)
    def exponential(self, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "exponential", scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def f(self, dfnum, dfden, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "f", dfnum, dfden, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def gamma(self, shape, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "gamma", shape, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def geometric(self, p, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "geometric", p, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def gumbel(self, loc=0.0, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "gumbel", loc, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def hypergeometric(self, ngood, nbad, nsample, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self,
            "hypergeometric",
            ngood,
            nbad,
            nsample,
            size=size,
            chunks=chunks,
            **kwargs,
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def laplace(self, loc=0.0, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "laplace", loc, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def logistic(self, loc=0.0, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "logistic", loc, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def lognormal(self, mean=0.0, sigma=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "lognormal", mean, sigma, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def logseries(self, p, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "logseries", p, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def multinomial(self, n, pvals, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self,
            "multinomial",
            n,
            pvals,
            size=size,
            chunks=chunks,
            extra_chunks=((len(pvals),),),
            **kwargs,
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def negative_binomial(self, n, p, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "negative_binomial", n, p, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def noncentral_chisquare(self, df, nonc, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "noncentral_chisquare", df, nonc, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def noncentral_f(self, dfnum, dfden, nonc, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "noncentral_f", dfnum, dfden, nonc, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def normal(self, loc=0.0, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "normal", loc, scale, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def pareto(self, a, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "pareto", a, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def permutation(self, x):
        from dask.array.slicing import shuffle_slice

        if isinstance(x, numbers.Number):
            x = arange(x, chunks="auto")

        index = np.arange(len(x))
        self._numpy_state.shuffle(index)
        return shuffle_slice(x, index)

    @derived_from(np.random.RandomState, skipblocks=1)
    def poisson(self, lam=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "poisson", lam, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def power(self, a, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "power", a, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def randint(self, low, high=None, size=None, chunks="auto", dtype="l", **kwargs):
        return _wrap_func(
            self, "randint", low, high, size=size, chunks=chunks, dtype=dtype, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def random_integers(self, low, high=None, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "random_integers", low, high, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def random_sample(self, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "random_sample", size=size, chunks=chunks, **kwargs)

    random = random_sample

    @derived_from(np.random.RandomState, skipblocks=1)
    def rayleigh(self, scale=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "rayleigh", scale, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def standard_cauchy(self, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "standard_cauchy", size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def standard_exponential(self, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "standard_exponential", size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def standard_gamma(self, shape, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "standard_gamma", shape, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def standard_normal(self, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "standard_normal", size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def standard_t(self, df, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "standard_t", df, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def tomaxint(self, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "tomaxint", size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def triangular(self, left, mode, right, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "triangular", left, mode, right, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def uniform(self, low=0.0, high=1.0, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "uniform", low, high, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def vonmises(self, mu, kappa, size=None, chunks="auto", **kwargs):
        return _wrap_func(
            self, "vonmises", mu, kappa, size=size, chunks=chunks, **kwargs
        )

    @derived_from(np.random.RandomState, skipblocks=1)
    def wald(self, mean, scale, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "wald", mean, scale, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def weibull(self, a, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "weibull", a, size=size, chunks=chunks, **kwargs)

    @derived_from(np.random.RandomState, skipblocks=1)
    def zipf(self, a, size=None, chunks="auto", **kwargs):
        return _wrap_func(self, "zipf", a, size=size, chunks=chunks, **kwargs)


def _rng_from_bitgen(bitgen):
    # Assumes typename(bitgen) starts with importable
    # library name (e.g. "numpy" or "cupy")
    backend_name = typename(bitgen).split(".")[0]
    backend_lib = importlib.import_module(backend_name)
    return backend_lib.random.default_rng(bitgen)


def _shuffle(bit_generator, x, axis=0):
    state_data = bit_generator.state
    bit_generator = type(bit_generator)()
    bit_generator.state = state_data
    state = _rng_from_bitgen(bit_generator)
    return state.shuffle(x, axis=axis)


def _spawn_bitgens(bitgen, n_bitgens):
    seeds = bitgen._seed_seq.spawn(n_bitgens)
    bitgens = [type(bitgen)(seed) for seed in seeds]
    return bitgens


def _apply_random_func(rng, funcname, bitgen, size, args, kwargs):
    """Apply random module method with seed"""
    if isinstance(bitgen, np.random.SeedSequence):
        bitgen = rng(bitgen)
    rng = _rng_from_bitgen(bitgen)
    func = getattr(rng, funcname)
    return func(*args, size=size, **kwargs)


def _apply_random(RandomState, funcname, state_data, size, args, kwargs):
    """Apply RandomState method with seed"""
    if RandomState is None:
        RandomState = array_creation_dispatch.RandomState
    state = RandomState(state_data)
    func = getattr(state, funcname)
    return func(*args, size=size, **kwargs)


def _choice_rng(state_data, a, size, replace, p, axis, shuffle):
    state = _rng_from_bitgen(state_data)
    return state.choice(a, size=size, replace=replace, p=p, axis=axis, shuffle=shuffle)


def _choice_rs(state_data, a, size, replace, p):
    state = array_creation_dispatch.RandomState(state_data)
    return state.choice(a, size=size, replace=replace, p=p)


def _choice_validate_params(state, a, size, replace, p, axis, chunks):
    dependencies = []
    # Normalize and validate `a`
    if isinstance(a, Integral):
        if isinstance(state, Generator):
            if state._backend_name == "cupy":
                raise NotImplementedError(
                    "`choice` not supported for cupy-backed `Generator`."
                )
            meta = state._backend.random.default_rng().choice(1, size=(), p=None)
        elif isinstance(state, RandomState):
            # On windows the output dtype differs if p is provided or
            # # absent, see https://github.com/numpy/numpy/issues/9867
            dummy_p = state._backend.array([1]) if p is not None else p
            meta = state._backend.random.RandomState().choice(1, size=(), p=dummy_p)
        else:
            raise ValueError("Unknown generator class")
        len_a = a
        if a < 0:
            raise ValueError("a must be greater than 0")
    else:
        a = asarray(a)
        a = a.rechunk(a.shape)
        meta = a._meta
        if a.ndim != 1:
            raise ValueError("a must be one dimensional")
        len_a = len(a)
        dependencies.append(a)
        a = a.__dask_keys__()[0]

    # Normalize and validate `p`
    if p is not None:
        if not isinstance(p, Array):
            # If p is not a dask array, first check the sum is close
            # to 1 before converting.
            p = asarray_safe(p, like=p)
            if not np.isclose(p.sum(), 1, rtol=1e-7, atol=0):
                raise ValueError("probabilities do not sum to 1")
            p = asarray(p)
        else:
            p = p.rechunk(p.shape)

        if p.ndim != 1:
            raise ValueError("p must be one dimensional")
        if len(p) != len_a:
            raise ValueError("a and p must have the same size")

        dependencies.append(p)
        p = p.__dask_keys__()[0]

    if size is None:
        size = ()

    if axis != 0:
        raise ValueError("axis must be 0 since a is one dimensional")

    chunks = normalize_chunks(chunks, size, dtype=np.float64)
    if not replace and len(chunks[0]) > 1:
        err_msg = (
            "replace=False is not currently supported for "
            "dask.array.choice with multi-chunk output "
            "arrays"
        )
        raise NotImplementedError(err_msg)

    return a, size, replace, p, axis, chunks, meta, dependencies


def _wrap_func(
    rng, funcname, *args, size=None, chunks="auto", extra_chunks=(), **kwargs
):
    if size is not None and not isinstance(size, (tuple, list)):
        size = (size,)
    return new_collection(
        Random(rng, funcname, size, chunks, extra_chunks, args, kwargs)
    )


class Random(IO):
    _parameters = [
        "rng",
        "distribution",
        "size",
        "chunks",
        "extra_chunks",
        "args",
        "kwargs",
    ]
    _defaults = {"extra_chunks": ()}

    @cached_property
    def kwargs(self):
        return self.operand("kwargs")

    @property
    def chunks(self):
        size = self.operand("size")
        chunks = self.operand("chunks")

        # shapes = list(
        #     {
        #         ar.shape
        #         for ar in chain(args, kwargs.values())
        #         if isinstance(ar, (Array, np.ndarray))
        #     }
        # )
        # if size is not None:
        #     shapes.append(size)
        shapes = [size]
        # broadcast to the final size(shape)
        size = broadcast_shapes(*shapes)
        return normalize_chunks(
            chunks,
            size,  # ideally would use dtype here
            dtype=self.kwargs.get("dtype", np.float64),
        )

    @cached_property
    def _info(self):
        sizes = list(product(*self.chunks))
        if isinstance(self.rng, Generator):
            bitgens = _spawn_bitgens(self.rng._bit_generator, len(sizes))
            bitgen_token = tokenize(bitgens)
            bitgens = [_bitgen._seed_seq for _bitgen in bitgens]
            func_applier = _apply_random_func
            gen = type(self.rng._bit_generator)
        elif isinstance(self.rng, RandomState):
            bitgens = random_state_data(len(sizes), self.rng._numpy_state)
            bitgen_token = tokenize(bitgens)
            func_applier = _apply_random
            gen = self.rng._RandomState
        else:
            raise TypeError(
                "Unknown object type: Not a Generator and Not a RandomState"
            )
        token = tokenize(bitgen_token, self.size, self.chunks, self.args, self.kwargs)
        name = f"{self.distribution}-{token}"

        return bitgens, name, sizes, gen, func_applier

    @property
    def _name(self):
        return self._info[1]

    @property
    def bitgens(self):
        return self._info[0]

    def _layer(self):
        bitgens, name, sizes, gen, func_applier = self._info

        keys = product(
            [name],
            *([range(len(bd)) for bd in self.chunks] + [[0]] * len(self.extra_chunks)),
        )

        vals = []
        # TODO: handle non-trivial args/kwargs (arrays, dask or otherwise)
        for bitgen, size in zip(bitgens, sizes):
            vals.append(
                (
                    func_applier,
                    gen,
                    self.distribution,
                    bitgen,
                    size,
                    self.args,
                    self.kwargs,
                )
            )

        return dict(zip(keys, vals))

    @cached_property
    def _meta(self):
        bitgens, name, sizes, gen, func_applier = self._info
        return func_applier(
            gen,
            self.distribution,
            bitgens[0],  # TODO: not sure about this
            (0,) * len(self.operand("size")),
            self.args,
            self.kwargs,
            # small_args,
            # small_kwargs,
        )


class RandomChoice(IO):
    _parameters = [
        "array",
        "chunks",
        "_meta",
        "_state",
        "replace",
        "p",
        "axis",
        "shuffle",
    ]
    _defaults = {"axis": None, "shuffle": None}
    _funcname = "da.random.choice-"

    @cached_property
    def chunks(self):
        return self.operand("chunks")

    @cached_property
    def sizes(self):
        return list(product(*self.chunks))

    @cached_property
    def state_data(self):
        return random_state_data(len(self.sizes), self._state)

    @cached_property
    def _meta(self):
        return self.operands("_meta")

    def _layer(self) -> dict:
        return {
            k: (_choice_rs, state, self.array, size, self.replace, self.p)
            for k, state, size in zip(self.__dask_keys__(), self.state_data, self.sizes)
        }


class RandomChoiceGenerator(RandomChoice):
    _defaults = {}

    @cached_property
    def state_data(self):
        return _spawn_bitgens(self._state, len(self.sizes))

    def _layer(self) -> dict:
        return {
            k: (
                _choice_rng,
                bitgen,
                self.array,
                size,
                self.replace,
                self.p,
                self.axis,
                self.shuffle,
            )
            for k, bitgen, size in zip(
                self.__dask_keys__(), self.state_data, self.sizes
            )
        }


"""
Lazy RNG-state machinery

Many of the RandomState methods are exported as functions in da.random for
backward compatibility reasons. Their usage is discouraged.
Use da.random.default_rng() to get a Generator based rng and use its
methods instead.
"""

_cached_states: dict[str, RandomState] = {}
_cached_states_lock = Lock()


def _make_api(attr):
    def wrapper(*args, **kwargs):
        key = array_creation_dispatch.backend
        with _cached_states_lock:
            try:
                state = _cached_states[key]
            except KeyError:
                _cached_states[key] = state = RandomState()
        return getattr(state, attr)(*args, **kwargs)

    wrapper.__name__ = getattr(RandomState, attr).__name__
    wrapper.__doc__ = getattr(RandomState, attr).__doc__
    return wrapper


"""
RandomState only
"""

seed = _make_api("seed")

beta = _make_api("beta")
binomial = _make_api("binomial")
chisquare = _make_api("chisquare")
choice = _make_api("choice")
exponential = _make_api("exponential")
f = _make_api("f")
gamma = _make_api("gamma")
geometric = _make_api("geometric")
gumbel = _make_api("gumbel")
hypergeometric = _make_api("hypergeometric")
laplace = _make_api("laplace")
logistic = _make_api("logistic")
lognormal = _make_api("lognormal")
logseries = _make_api("logseries")
multinomial = _make_api("multinomial")
negative_binomial = _make_api("negative_binomial")
noncentral_chisquare = _make_api("noncentral_chisquare")
noncentral_f = _make_api("noncentral_f")
normal = _make_api("normal")
pareto = _make_api("pareto")
permutation = _make_api("permutation")
poisson = _make_api("poisson")
power = _make_api("power")
random_sample = _make_api("random_sample")
random = _make_api("random_sample")
randint = _make_api("randint")
random_integers = _make_api("random_integers")
rayleigh = _make_api("rayleigh")
standard_cauchy = _make_api("standard_cauchy")
standard_exponential = _make_api("standard_exponential")
standard_gamma = _make_api("standard_gamma")
standard_normal = _make_api("standard_normal")
standard_t = _make_api("standard_t")
triangular = _make_api("triangular")
uniform = _make_api("uniform")
vonmises = _make_api("vonmises")
wald = _make_api("wald")
weibull = _make_api("weibull")
zipf = _make_api("zipf")



from __future__ import annotations

import numpy as np
import pytest

import dask
from dask.dataframe.dask_expr import Repartition, from_pandas, map_partitions, merge
from dask.dataframe.dask_expr._merge import BroadcastJoin
from dask.dataframe.dask_expr._shuffle import P2PShuffle, TaskShuffle
from dask.dataframe.dask_expr.tests._util import _backend_library

distributed = pytest.importorskip("distributed")

from distributed import Client, LocalCluster
from distributed.shuffle._core import id_from_key
from distributed.utils_test import gen_cluster  # noqa F401

import dask.dataframe as dd

# Set DataFrame backend for this module
pd = _backend_library()


@pytest.fixture
def pdf():
    return pd.DataFrame({"x": list(range(20)) * 5, "y": range(100)})


@pytest.fixture
def df(pdf):
    return from_pandas(pdf, npartitions=10)


@pytest.mark.parametrize("npartitions", [None, 1, 20])
@gen_cluster(client=True)
async def test_p2p_shuffle(c, s, a, b, npartitions):
    df = dd.dask_expr.datasets.timeseries(
        start="2000-01-01",
        end="2000-01-10",
        dtypes={"x": float, "y": float},
        freq="10 s",
    )
    out = df.shuffle("x", shuffle_method="p2p", npartitions=npartitions)
    if npartitions is None:
        assert out.npartitions == df.npartitions
    else:
        assert out.npartitions == npartitions
    x, y, z = c.compute([df.x.size, out.x.size, out.partitions[-1].x.size])
    x = await x
    y = await y
    z = await z
    assert x == y
    if npartitions != 1:
        assert x > z


@pytest.mark.parametrize("npartitions_left", [5, 6])
@gen_cluster(client=True)
async def test_merge_p2p_shuffle(c, s, a, b, npartitions_left):
    df_left = pd.DataFrame({"a": [1, 2, 3] * 100, "b": 2})
    df_right = pd.DataFrame({"a": [4, 2, 3] * 100, "c": 2})
    left = from_pandas(df_left, npartitions=npartitions_left)
    right = from_pandas(df_right, npartitions=5)

    out = left.merge(right, shuffle_method="p2p")
    assert out.npartitions == npartitions_left
    x = c.compute(out)
    x = await x
    pd.testing.assert_frame_equal(x.reset_index(drop=True), df_left.merge(df_right))


@gen_cluster(client=True)
async def test_self_merge_p2p_shuffle(c, s, a, b):
    pdf = pd.DataFrame({"a": range(100), "b": range(0, 200, 2)})
    ddf = from_pandas(pdf, npartitions=5)

    out = ddf.merge(ddf, left_on="a", right_on="b", shuffle_method="p2p")
    # Generate unique shuffle IDs if the input frame is the same but parameters differ
    assert sum(id_from_key(k) is not None for k in out.dask) == 2
    x = await c.compute(out)
    expected = pdf.merge(pdf, left_on="a", right_on="b")
    pd.testing.assert_frame_equal(
        x.sort_values("a_x", ignore_index=True),
        expected.sort_values("a_x", ignore_index=True),
    )


@gen_cluster(client=True)
@pytest.mark.parametrize("name", ["a", None])
@pytest.mark.parametrize("shuffle", ["tasks", "disk", "p2p"])
async def test_merge_index_precedence(c, s, a, b, shuffle, name):
    pdf = pd.DataFrame(
        {"a": [1, 2, 3, 4, 5, 6]}, index=pd.Index([6, 5, 4, 3, 2, 1], name=name)
    )
    pdf2 = pd.DataFrame(
        {"b": [1, 2, 3, 4, 5, 6]}, index=pd.Index([1, 2, 7, 4, 5, 6], name=name)
    )
    df = from_pandas(pdf, npartitions=2, sort=False)
    df2 = from_pandas(pdf2, npartitions=3, sort=False)

    result = df.join(df2, shuffle_method=shuffle)
    x = await c.compute(result)
    assert result.npartitions == 3
    pd.testing.assert_frame_equal(x.sort_index(ascending=False), pdf.join(pdf2))


@pytest.mark.filterwarnings("ignore:DataFrameGroupBy.apply operated on")
@gen_cluster(client=True)
@pytest.mark.parametrize("shuffle", ["disk", "p2p", "tasks"])
async def test_shuffle_nulls_introduced(c, s, a, b, shuffle):
    df1 = pd.DataFrame([[True], [False]] * 50, columns=["A"])
    df1["B"] = list(range(100))

    df2 = pd.DataFrame(
        [[2, 3], [109, 2], [345, 3], [50, 7], [95, 1]], columns=["B", "C"]
    )

    ddf1 = from_pandas(df1, npartitions=10)
    ddf2 = from_pandas(df2, npartitions=1)
    meta = pd.Series(dtype=int, index=pd.Index([], dtype=bool, name="A"), name="A")
    result = (
        merge(ddf1, ddf2, how="outer", on="B", shuffle_method=shuffle)
        .groupby("A")
        .apply(lambda df: len(df), meta=meta, shuffle_method=shuffle)
    )
    expected = (
        pd.merge(df1, df2, how="outer", on="B").groupby("A").apply(lambda df: len(df))
    )
    x = await c.compute(result)

    pd.testing.assert_series_equal(
        x.sort_index(), expected.sort_index(), check_names=False
    )


@gen_cluster(client=True)
@pytest.mark.parametrize("shuffle", ["tasks", "p2p"])
@pytest.mark.parametrize("broadcast", [True, 0.6])
@pytest.mark.parametrize("how", ["left", "inner"])
async def test_merge_broadcast(c, s, a, b, shuffle, broadcast, how):
    pdf = pd.DataFrame({"a": [1, 2, 3, 4, 5, 6] * 5, "c": 1})
    pdf2 = pd.DataFrame({"b": [1, 2, 3, 4, 5, 6]})
    df = from_pandas(pdf, npartitions=15)
    df2 = from_pandas(pdf2, npartitions=2)

    result = df.merge(
        df2,
        left_on="a",
        right_on="b",
        shuffle_method=shuffle,
        broadcast=broadcast,
        how=how,
    )
    q = result.optimize()
    assert len(list(q.find_operations(BroadcastJoin))) > 0
    x = await c.compute(result)
    assert result.npartitions == 15
    pd.testing.assert_frame_equal(
        x.sort_values(by="a", ignore_index=True),
        pdf.merge(pdf2, left_on="a", right_on="b", how=how).sort_values(
            by="a", ignore_index=True
        ),
    )


@gen_cluster(client=True)
async def test_merge_p2p_shuffle_reused_dataframe_with_different_parameters(c, s, a, b):
    pdf1 = pd.DataFrame({"a": range(100), "b": range(0, 200, 2)})
    pdf2 = pd.DataFrame({"x": range(200), "y": [1, 2, 3, 4] * 50})
    ddf1 = from_pandas(pdf1, npartitions=5)
    ddf2 = from_pandas(pdf2, npartitions=10)

    out = (
        ddf1.merge(ddf2, left_on="a", right_on="x", shuffle_method="p2p")
        # Vary the number of output partitions for the shuffles of dd2
        .repartition(npartitions=20).merge(
            ddf2, left_on="b", right_on="x", shuffle_method="p2p"
        )
    )
    # Generate unique shuffle IDs if the input frame is the same but
    # parameters differ. Reusing shuffles in merges is dangerous because of the
    # required coordination and complexity introduced through dynamic clusters.
    assert sum(id_from_key(k) is not None for k in out.dask) == 4
    x = await c.compute(out)
    expected = pdf1.merge(pdf2, left_on="a", right_on="x").merge(
        pdf2, left_on="b", right_on="x"
    )
    pd.testing.assert_frame_equal(
        x.sort_values("a", ignore_index=True),
        expected.sort_values("a", ignore_index=True),
    )


@gen_cluster(client=True)
async def test_merge_p2p_shuffle_reused_dataframe_with_same_parameters(c, s, a, b):
    pdf1 = pd.DataFrame({"a": range(100), "b": range(0, 200, 2)})
    pdf2 = pd.DataFrame({"x": range(200), "y": [1, 2, 3, 4] * 50})
    ddf1 = from_pandas(pdf1, npartitions=5)
    ddf2 = from_pandas(pdf2, npartitions=10)

    # This performs two shuffles:
    #   * ddf1 is shuffled on `a`
    #   * ddf2 is shuffled on `x`
    ddf3 = ddf1.merge(ddf2, left_on="a", right_on="x", shuffle_method="p2p")

    # This performs one shuffle:
    #   * ddf3 is shuffled on `b`
    # We can reuse the shuffle of dd2 on `x` from the previous merge.
    out = ddf2.merge(
        ddf3,
        left_on="x",
        right_on="b",
        shuffle_method="p2p",
    )
    # Generate unique shuffle IDs if the input frame is the same and all its
    # parameters match. Reusing shuffles in merges is dangerous because of the
    # required coordination and complexity introduced through dynamic clusters.
    assert sum(id_from_key(k) is not None for k in out.dask) == 4
    x = await c.compute(out)
    expected = pdf2.merge(
        pdf1.merge(pdf2, left_on="a", right_on="x"), left_on="x", right_on="b"
    )
    pd.testing.assert_frame_equal(
        x.sort_values("a", ignore_index=True),
        expected.sort_values("a", ignore_index=True),
    )


@pytest.mark.parametrize("npartitions_left", [5, 6])
@gen_cluster(client=True)
async def test_index_merge_p2p_shuffle(c, s, a, b, npartitions_left):
    df_left = pd.DataFrame({"a": [1, 2, 3] * 100, "b": 2}).set_index("a")
    df_right = pd.DataFrame({"a": [4, 2, 3] * 100, "c": 2})
    left = from_pandas(df_left, npartitions=npartitions_left, sort=False)
    right = from_pandas(df_right, npartitions=5)

    out = left.merge(right, left_index=True, right_on="a", shuffle_method="p2p")
    assert out.npartitions == npartitions_left
    x = c.compute(out)
    x = await x
    pd.testing.assert_frame_equal(
        x.sort_index(),
        df_left.merge(df_right, left_index=True, right_on="a").sort_index(),
    )


@gen_cluster(client=True)
async def test_merge_p2p_shuffle(c, s, a, b):
    df_left = pd.DataFrame({"a": [1, 2, 3] * 100, "b": 2, "e": 2})
    df_right = pd.DataFrame({"a": [4, 2, 3] * 100, "c": 2})
    left = from_pandas(df_left, npartitions=6)
    right = from_pandas(df_right, npartitions=5)

    out = left.merge(right, shuffle_method="p2p")[["b", "c"]]
    assert out.npartitions == 6
    x = c.compute(out)
    x = await x
    pd.testing.assert_frame_equal(
        x.reset_index(drop=True), df_left.merge(df_right)[["b", "c"]]
    )


@gen_cluster(client=True)
async def test_merge_p2p_shuffle_projection_error(c, s, a, b):
    pdf1 = pd.DataFrame({"a": [1, 2, 3], "b": 1})
    pdf2 = pd.DataFrame({"x": [1, 2, 3, 4, 5, 6], "y": 1})
    df1 = from_pandas(pdf1, npartitions=2)
    df2 = from_pandas(pdf2, npartitions=3)
    df = df1.merge(df2, left_on="a", right_on="x")
    min_val = df.groupby("x")["y"].sum().reset_index()
    result = df.merge(min_val)
    expected = pd.DataFrame(
        {"a": [2, 3, 1], "b": 1, "x": [2, 3, 1], "y": 1}, index=[0, 0, 1]
    )
    x = c.compute(result)
    x = await x
    pd.testing.assert_frame_equal(
        x.sort_values("a", ignore_index=True),
        expected.sort_values("a", ignore_index=True),
    )


def test_sort_values():
    with LocalCluster(processes=False, n_workers=2, dashboard_address=":0") as cluster:
        with Client(cluster) as client:  # noqa: F841
            pdf = pd.DataFrame({"a": [5] + list(range(100)), "b": 2})
            df = from_pandas(pdf, npartitions=10)

            out = df.sort_values(by="a").compute()
    pd.testing.assert_frame_equal(
        out.reset_index(drop=True),
        pdf.sort_values(by="a", ignore_index=True),
    )


@pytest.mark.parametrize("add_repartition", [True, False])
def test_merge_combine_similar_squash_merges(add_repartition):
    with LocalCluster(processes=False, n_workers=2, dashboard_address=":0") as cluster:
        with Client(cluster) as client:  # noqa: F841
            pdf = pd.DataFrame(
                {
                    "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] * 2,
                    "b": 1,
                    "c": 1,
                }
            )
            pdf2 = pd.DataFrame(
                {"m": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] * 2, "n": 1, "o": 2, "p": 3}
            )
            pdf3 = pd.DataFrame(
                {"x": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] * 2, "y": 1, "z": 1, "zz": 2}
            )

            df = from_pandas(pdf, npartitions=5)
            df2 = from_pandas(pdf2, npartitions=10)
            df3 = from_pandas(pdf3, npartitions=10)

            df = df[df.a > 1]
            df2 = df2[df2.m > 1]
            df3 = df3[df3.x > 1]
            if add_repartition:
                df = df.repartition(npartitions=df.npartitions // 2)
                df2 = df2.repartition(npartitions=df2.npartitions // 2)
            q = df.merge(df2, left_on="a", right_on="m")
            if add_repartition:
                df3 = df3.repartition(npartitions=df3.npartitions // 2)
                q = q.repartition(npartitions=q.npartitions // 2)
            q = q.merge(df3, left_on="n", right_on="x")
            q["revenue"] = q.y * (1 - q.z)
            result = q[["x", "n", "o", "revenue"]]
            result_q = result.optimize(fuse=False)

            assert (
                result_q.expr.frame.frame.frame._name
                == result_q.expr.frame.operands[2].left.frame._name
            )
            out = result.compute()

    pdf = pdf[pdf.a > 1]
    pdf2 = pdf2[pdf2.m > 1]
    pdf3 = pdf3[pdf3.x > 1]
    q = pdf.merge(pdf2, left_on="a", right_on="m")
    q = q.merge(pdf3, left_on="n", right_on="x")
    q["revenue"] = q.y * (1 - q.z)
    expected = q[["x", "n", "o", "revenue"]]

    pd.testing.assert_frame_equal(
        out.reset_index(drop=True),
        expected,
    )


@pytest.mark.parametrize("kind", ["future", "delayed"])
@gen_cluster(client=True)
async def test_future_in_map_partitions(c, s, a, b, kind):
    # xgboost uses this pattern

    def test_func(n):
        import pandas as pd

        return pd.DataFrame({"a": list(range(n))})

    df = from_pandas(pd.DataFrame({"a": [1, 2, 3, 4]}), npartitions=2)

    if kind == "future":
        f = c.submit(test_func, 100)
        z = c.submit(test_func, 10)
    elif kind == "delayed":
        f = dask.delayed(test_func)(100)
        z = dask.delayed(test_func)(10)
    else:
        raise ValueError("kind must be 'future' or 'delayed'")

    def _foo(x, y, z):
        return x.sum() + y + z.sum()

    q = map_partitions(_foo, f, df, z=z, meta=df._meta)
    result = c.compute(q)
    result = await result
    expected = pd.DataFrame({"a": [4996, 4997, 4998, 4999]})
    pd.testing.assert_frame_equal(result, expected)


@gen_cluster(client=True)
async def test_shuffle_consistency_checks(c, s, a, b):
    pdf = pd.DataFrame({"x": [1, 2, 3]})
    df = from_pandas(pdf, npartitions=2)
    df2 = df.astype(np.csingle)
    with pytest.raises(TypeError, match="p2p does not support data of type"):
        df2.shuffle("x")

    df.columns = [1]
    with pytest.raises(TypeError, match="p2p requires all column names to be str"):
        df.shuffle(1)


@gen_cluster(client=True)
async def test_merge_indicator(c, s, a, b):
    data = {
        "id": ["101-a", "102-a", "103-a"],
        "test": ["val101a", "val102a", "val103a"],
    }
    pdf = pd.DataFrame(data, dtype="string[pyarrow]")
    df = from_pandas(pdf, npartitions=2)
    result = df.merge(df, on="id", how="outer", indicator=True)
    x = c.compute(result)
    x = await x
    expected = pdf.merge(pdf, on="id", how="outer", indicator=True)

    pd.testing.assert_frame_equal(
        x.sort_values("id", ignore_index=True),
        expected.sort_values("id", ignore_index=True),
    )


@gen_cluster(client=True)
async def test_shuffle_partition_reduction(c, s, a, b):
    pdf = pd.DataFrame({"a": [1, 2, 3, 4] * 100, "b": 1})
    df = from_pandas(pdf, npartitions=10)
    result = df.shuffle(on="a", npartitions=4)
    q = result.optimize(fuse=False)
    assert not any(isinstance(op, Repartition) for op in q.walk())
    x = c.compute(result)
    x = await x

    pd.testing.assert_frame_equal(
        x.sort_values("a", ignore_index=True),
        pdf.sort_values("a", ignore_index=True),
    )


@gen_cluster(client=True)
async def test_p2p_and_merge_shuffle(c, s, a, b):
    pdf = pd.DataFrame({"a": np.random.randint(1, 100, (100,)), "b": 1})

    df = from_pandas(pdf, npartitions=15)

    pdf2 = pd.DataFrame({"a": np.random.randint(1, 100, (100,)), "c": 1})
    df2 = from_pandas(pdf2, npartitions=10)

    pdf3 = pd.DataFrame({"a": np.random.randint(1, 100, (100,)), "d": 2})
    df3 = from_pandas(pdf3, npartitions=5)

    result = df.merge(df2)
    result = result.merge(df3)
    x = c.compute(result)
    x = await x
    pd.testing.assert_frame_equal(
        x.sort_values("a", ignore_index=True),
        pdf.merge(pdf2).merge(pdf3).sort_values("a", ignore_index=True),
    )


@pytest.mark.parametrize("func", ["set_index", "sort_values", "shuffle"])
def test_respect_context_shuffle(df, pdf, func):
    with dask.config.set({"dataframe.shuffle.method": "tasks"}):
        q = getattr(df, func)("x")
    result = q.optimize(fuse=False)
    assert len([x for x in result.walk() if isinstance(x, TaskShuffle)]) > 0

    q = getattr(df, func)("x")
    with dask.config.set({"dataframe.shuffle.method": "tasks"}):
        result = q.optimize(fuse=False)
    assert len([x for x in result.walk() if isinstance(x, TaskShuffle)]) > 0

    with dask.config.set({"dataframe.shuffle.method": "p2p"}):
        q = getattr(df, func)("x")
    with dask.config.set({"dataframe.shuffle.method": "tasks"}):
        result = q.optimize(fuse=False)
    assert len([x for x in result.walk() if isinstance(x, P2PShuffle)]) > 0



from __future__ import annotations

from collections import defaultdict

from dask.dataframe import methods
from dask.dataframe.dispatch import (  # noqa: F401
    categorical_dtype,
    categorical_dtype_dispatch,
    is_categorical_dtype,
)


def _categorize_block(df, categories, index):
    """Categorize a dataframe with given categories

    df: DataFrame
    categories: dict mapping column name to iterable of categories
    """
    df = df.copy()
    for col, vals in categories.items():
        if is_categorical_dtype(df[col]):
            df[col] = df[col].cat.set_categories(vals)
        else:
            cat_dtype = categorical_dtype(meta=df[col], categories=vals, ordered=False)
            df[col] = df[col].astype(cat_dtype)
    if index is not None:
        if is_categorical_dtype(df.index):
            ind = df.index.set_categories(index)
        else:
            cat_dtype = categorical_dtype(
                meta=df.index, categories=index, ordered=False
            )
            ind = df.index.astype(dtype=cat_dtype)
        ind.name = df.index.name
        df.index = ind
    return df


def _get_categories(df, columns, index):
    res = {}
    for col in columns:
        x = df[col]
        if is_categorical_dtype(x):
            res[col] = x._constructor(x.cat.categories)
        else:
            res[col] = x.dropna().drop_duplicates()
    if index:
        if is_categorical_dtype(df.index):
            return res, df.index.categories
        return res, df.index.dropna().drop_duplicates()
    return res, None


def _get_categories_agg(parts):
    res = defaultdict(list)
    res_ind = []
    for p in parts:
        for k, v in p[0].items():
            res[k].append(v)
        res_ind.append(p[1])
    res = {
        k: methods.concat(v, ignore_index=True).drop_duplicates()
        for k, v in res.items()
    }
    if res_ind[0] is None:
        return res, None
    return res, res_ind[0].append(res_ind[1:]).drop_duplicates()



"""Tests on traceback shortening heuristics

See Also
--------
distributed/tests/test_client.py::test_short_tracebacks
distributed/tests/test_client.py::test_short_tracebacks_async
"""

from __future__ import annotations

import traceback
from contextlib import contextmanager

import pytest

import dask
from dask.utils import shorten_traceback


@contextmanager
def assert_tb_levels(expect):
    with pytest.raises(ZeroDivisionError) as e:
        yield
    frames = list(traceback.walk_tb(e.tb))
    frame_names = [frame[0].f_code.co_name for frame in frames]
    assert frame_names[0] == "assert_tb_levels", frame_names
    assert frame_names[1:] == expect, frame_names


def f1():
    return 1 / 0


def f2():
    return f1() + 5


def f3():
    return f2() + 1


@pytest.mark.parametrize(
    "regexes,expect",
    [
        (None, ["test_shorten_traceback", "f3", "f2", "f1"]),  # Disabled
        ([], ["test_shorten_traceback", "f3", "f2", "f1"]),  # Disabled
        (["nomatch"], ["test_shorten_traceback", "f3", "f2", "f1"]),
        # Remove absolutely everything, but keep the first and last frame
        # Python <= 3.11 can't remove the first frame of a traceback from a __exit__
        # function
        ([".*"], ["test_shorten_traceback", "f1"]),
        (["tests"], ["test_shorten_traceback", "f1"]),
    ],
)
def test_shorten_traceback(regexes, expect):
    """
    See also
    --------
    test_distributed.py::test_shorten_traceback_excepthook
    test_distributed.py::test_shorten_traceback_ipython
    """
    with dask.config.set({"admin.traceback.shorten": regexes}):
        with assert_tb_levels(expect):
            with shorten_traceback():
                f3()


try:
    import tblib
except ImportError:
    tblib = None


@pytest.mark.parametrize("scheduler", ["threads", "processes", "sync"])
def test_compute_shorten_traceback(scheduler):
    d = dask.delayed(f3)()

    TEST_NAME = "test_compute_shorten_traceback"
    if scheduler == "processes" and not tblib:
        remote_stack = ["reraise"]
    else:
        remote_stack = ["f3", "f2", "f1"]

    expect = [TEST_NAME, "compute", *remote_stack]
    with assert_tb_levels(expect):
        dask.compute(d(), scheduler=scheduler)

    expect = [TEST_NAME, "compute", "compute", *remote_stack]
    with assert_tb_levels(expect):
        d.compute(scheduler=scheduler)


@pytest.mark.parametrize("scheduler", ["threads", "processes", "sync"])
def test_persist_shorten_traceback(scheduler):
    d = dask.delayed(f3)()

    TEST_NAME = "test_persist_shorten_traceback"
    if scheduler == "processes" and not tblib:
        remote_stack = ["reraise"]
    else:
        remote_stack = ["f3", "f2", "f1"]

    expect = [TEST_NAME, "persist", *remote_stack]
    with assert_tb_levels(expect):
        dask.persist(d(), scheduler=scheduler)

    expect = [TEST_NAME, "persist", "persist", *remote_stack]
    with assert_tb_levels(expect):
        d.persist(scheduler=scheduler)


def test_distributed_shorten_traceback():
    distributed = pytest.importorskip("distributed")
    with distributed.Client(processes=False, dashboard_address=":0"):
        d = dask.delayed(f3)()

        (dp1,) = dask.persist(d)
        dp2 = d.persist()

        TEST_NAME = "test_distributed_shorten_traceback"
        expect = [TEST_NAME, "compute", "f3", "f2", "f1"]
        with assert_tb_levels(expect):
            dask.compute(d())

        expect = [TEST_NAME, "compute", "compute", "f3", "f2", "f1"]
        with assert_tb_levels(expect):
            d.compute()

        expect = [TEST_NAME, "compute", "f3", "f2", "f1"]
        with assert_tb_levels(expect):
            dask.compute(dp1)

        expect = [TEST_NAME, "compute", "compute", "f3", "f2", "f1"]
        with assert_tb_levels(expect):
            dp2.compute()


def test_deprecated_config(tmp_path):
    """Test config override in the format between 2023.6.1 and 2023.8.1"""
    d = {}
    dask.config.refresh(config=d)
    actual = dask.config.get("admin.traceback.shorten", config=d)
    assert isinstance(actual, list) and len(actual) > 2

    d = {}
    with open(tmp_path / "dask.yaml", "w") as fh:
        fh.write(
            """
            admin:
              traceback:
                shorten:
                  when:
                    - dask/base.py
                  what:
                    - dask/core.py
            """
        )

    with pytest.warns(FutureWarning):
        dask.config.refresh(config=d, paths=[tmp_path])
    actual = dask.config.get("admin.traceback.shorten", config=d)
    assert actual == ["dask/core.py"]



from __future__ import annotations

import datetime

CombinedOutput = type("CombinedOutput", (tuple,), {})


def overlap_chunk(func, before, after, *args, **kwargs):
    dfs = [df for df in args if isinstance(df, CombinedOutput)]
    combined, prev_part_length, next_part_length = dfs[0]

    args = [arg[0] if isinstance(arg, CombinedOutput) else arg for arg in args]

    out = func(*args, **kwargs)

    if prev_part_length is None:
        before = None
    if isinstance(before, datetime.timedelta):
        before = prev_part_length

    expansion = None
    if combined.shape[0] != 0:
        expansion = out.shape[0] // combined.shape[0]
    if before and expansion:
        before *= expansion
    if next_part_length is None:
        return out.iloc[before:]
    if isinstance(after, datetime.timedelta):
        after = next_part_length
    if after and expansion:
        after *= expansion
    return out.iloc[before:-after]


def _head_timedelta(current, next_, after):
    """Return rows of ``next_`` whose index is before the last
    observation in ``current`` + ``after``.

    Parameters
    ----------
    current : DataFrame
    next_ : DataFrame
    after : timedelta

    Returns
    -------
    overlapped : DataFrame
    """
    return next_[next_.index < (current.index.max() + after)]



from __future__ import annotations

from dask.dataframe.dask_expr._accessor import Accessor


class DatetimeAccessor(Accessor):
    """Accessor object for datetimelike properties of the Series values.

    Examples
    --------

    >>> s.dt.microsecond  # doctest: +SKIP
    """

    _accessor_name = "dt"

    _accessor_methods = (
        "ceil",
        "day_name",
        "floor",
        "isocalendar",
        "month_name",
        "normalize",
        "round",
        "strftime",
        "to_period",
        "to_pydatetime",
        "to_pytimedelta",
        "to_timestamp",
        "total_seconds",
        "tz_convert",
        "tz_localize",
    )

    _accessor_properties = (
        "components",
        "date",
        "day",
        "day_of_week",
        "day_of_year",
        "dayofweek",
        "dayofyear",
        "days",
        "days_in_month",
        "daysinmonth",
        "end_time",
        "freq",
        "hour",
        "is_leap_year",
        "is_month_end",
        "is_month_start",
        "is_quarter_end",
        "is_quarter_start",
        "is_year_end",
        "is_year_start",
        "microsecond",
        "microseconds",
        "minute",
        "month",
        "nanosecond",
        "nanoseconds",
        "quarter",
        "qyear",
        "second",
        "seconds",
        "start_time",
        "time",
        "timetz",
        "tz",
        "week",
        "weekday",
        "weekofyear",
        "year",
    )



from __future__ import annotations

from dask.dataframe.dask_expr import datasets
from dask.dataframe.dask_expr._collection import *
from dask.dataframe.dask_expr._dummies import get_dummies
from dask.dataframe.dask_expr._groupby import Aggregation
from dask.dataframe.dask_expr.io._delayed import from_delayed
from dask.dataframe.dask_expr.io.bag import to_bag
from dask.dataframe.dask_expr.io.parquet import to_parquet
from dask.dataframe.dask_expr.io.records import to_records



from __future__ import annotations

import pickle

import pytest

from dask._expr import HLGExpr, _ExprSequence
from dask._task_spec import (
    DataNode,
    DependenciesMapping,
    Task,
    TaskRef,
    fuse_linear_task_spec,
    resolve_aliases,
)
from dask.blockwise import blockwise, optimize_blockwise
from dask.core import flatten, reverse_dict
from dask.highlevelgraph import HighLevelGraph, MaterializedLayer
from dask.tokenize import tokenize
from dask.utils import ensure_dict


def test_tokenize():
    # Ensure tokens are different for different high-level graphs The current
    # implementation actually ensures that no HLGExpr are tokenizing equally.
    # Technically, we do not need such a strong guarantee. but tokenizing a full
    # HLG reliably is tricky and we do not require the reproducibility for
    # HLGExpr since they do not undergo the same kind of optimization as the
    # rest of the graph.
    from dask.highlevelgraph import HighLevelGraph

    dsk = HighLevelGraph.from_collections("x", {"foo": None})
    dsk2 = HighLevelGraph.from_collections("x", {"bar": None})
    dsk3 = HighLevelGraph.from_collections("y", {"foo": None})
    assert tokenize(HLGExpr(dsk)) != tokenize(HLGExpr(dsk2))
    assert tokenize(HLGExpr(dsk)) != tokenize(HLGExpr(dsk3))
    assert tokenize(HLGExpr(dsk2)) != tokenize(HLGExpr(dsk3))

    # Roundtrip preserves the tokens
    for expr in [HLGExpr(dsk), HLGExpr(dsk2), HLGExpr(dsk3)]:
        assert tokenize(pickle.loads(pickle.dumps(expr))) == tokenize(expr)


def optimizer(dsk, keys):
    dsk = ensure_dict(dsk)
    keys = list(flatten(keys))
    dsk = fuse_linear_task_spec(dsk, keys)
    return resolve_aliases(
        dsk,
        keys,
        reverse_dict(DependenciesMapping(dsk)),
    )


def optimizer2(dsk, keys):
    return optimizer(dsk, keys)


def func(*args, **kwargs):
    pass


def test_hlg_expr_sequence_finalize():
    hlgx = HighLevelGraph(
        xlayer := {"xlayer": MaterializedLayer({"x": DataNode("x", 1)})},
        dependencies=(xdeps := {"xlayer": set()}),
    )
    ylayer = {"ylayer": MaterializedLayer({"y": Task("y", func, TaskRef("x"))})}
    ylayer.update(xlayer)
    ydeps = {"ylayer": {"xlayer"}}
    ydeps.update(xdeps)
    hlgy = HighLevelGraph(ylayer, dependencies=ydeps)
    zlayer = {"zlayer": MaterializedLayer({"z": Task("z", func, TaskRef("x"))})}
    zlayer.update(xlayer)
    zdeps = {"zlayer": {"xlayer"}}

    zdeps.update(xdeps)
    hlgz = HighLevelGraph(zlayer, dependencies=zdeps)
    hlgexprx = HLGExpr(
        hlgx,
        low_level_optimizer=optimizer,
        output_keys=["x"],
    )
    hlgexpry = HLGExpr(
        hlgy,
        low_level_optimizer=optimizer,
        output_keys=["y"],
    )
    hlgexprz = HLGExpr(
        hlgz,
        low_level_optimizer=optimizer,
        output_keys=["z"],
    )
    dskx = hlgexprx.finalize_compute().optimize().__dask_graph__()
    assert isinstance(dskx, dict)
    assert len(dskx) == 1
    assert "x" in dskx
    assert dskx["x"] is hlgy.layers["xlayer"]["x"]

    dsky = hlgexpry.finalize_compute().optimize().__dask_graph__()
    assert isinstance(dsky, dict)
    # Linear low level fusion
    assert len(dsky) == 1
    assert "y" in dsky
    assert dsky["y"] != hlgy.layers["ylayer"]["y"]

    expryz_opt = _ExprSequence(hlgexprz, hlgexpry).finalize_compute().optimize()
    keys_yz = expryz_opt.__dask_keys__()
    assert len(keys_yz) == 2

    dskyz = expryz_opt.__dask_graph__()
    assert isinstance(dskyz, dict)
    expected = {}
    expected.update(next(iter(hlgx.layers.values())).mapping)
    expected.update(next(iter(hlgy.layers.values())).mapping)
    expected.update(next(iter(hlgz.layers.values())).mapping)
    # This is building the graph properly without fusing anything
    assert dskyz == expected

    hlgexprz_different_optimizer = HLGExpr(
        hlgz,
        low_level_optimizer=optimizer2,
        output_keys=["z"],
    )

    dskyz2 = (
        _ExprSequence(hlgexprz_different_optimizer, hlgexpry)
        .finalize_compute()
        .optimize()
        .__dask_graph__()
    )
    # both are fusing x
    assert "x" not in dskyz2
    assert len(dskyz2) == 2


def test_hlg_expr_sequence_nested_keys():
    xlayer = {"xlayer": MaterializedLayer({"x": DataNode("x", 1)})}
    xdeps = {"xlayer": set()}
    ylayer = {"ylayer": MaterializedLayer({"y": Task("y", func, TaskRef("x"))})}
    ylayer.update(xlayer)
    ydeps = {"ylayer": {"xlayer"}}
    ydeps.update(xdeps)
    hlgy = HighLevelGraph(ylayer, dependencies=ydeps)
    zlayer = {"zlayer": MaterializedLayer({"z": Task("z", func, TaskRef("x"))})}
    zlayer.update(xlayer)
    zdeps = {"zlayer": {"xlayer"}}

    zdeps.update(xdeps)
    hlgz = HighLevelGraph(zlayer, dependencies=zdeps)
    hlgexpry = HLGExpr(
        hlgy,
        low_level_optimizer=optimizer,
        output_keys=[["y"], ["x"]],
    )
    hlgexprz = HLGExpr(
        hlgz,
        low_level_optimizer=optimizer,
        output_keys=[["z"]],
    )
    expr = _ExprSequence(hlgexprz, hlgexpry)
    expected = [[["z"]], [["y"], ["x"]]]
    assert expr.__dask_keys__() == expected
    assert expr.optimize().__dask_keys__() == expected

    # Now with a different grouping / optimizer pass. These are handled
    # separately internalyl and we want to make sure the sequence is putting it
    # back together properly
    hlgexprz = HLGExpr(
        hlgz,
        low_level_optimizer=optimizer2,
        output_keys=[["z"]],
    )
    expr = _ExprSequence(hlgexprz, hlgexpry)
    assert expr.__dask_keys__() == expected
    assert expr.optimize().__dask_keys__() == expected


def optimizer_with_annotations(dsk, keys):
    annots = {}
    for layer in dsk.layers.values():
        if layer.annotations:
            annots.update(layer.annotations)
    out = optimizer(dsk, keys)
    out = {f"{key}-optimized": val for key, val in out.items()}
    return HighLevelGraph(
        layers={"fused": MaterializedLayer(out, annotations=annots)},
        dependencies={"fused": set()},
    )


def test_hlg_sequence_uses_annotations_of_optimized_dsk():
    """
    There was an issue where generating annotations would fall back to the HLG
    objects that were not optimized in combination.
    """
    xlayer = {
        "xlayer": MaterializedLayer({"x": DataNode("x", 1)}, annotations={"foo": 1})
    }
    xdeps = {"xlayer": set()}
    ylayer = {"ylayer": MaterializedLayer({"y": Task("y", func, TaskRef("x"))})}
    ylayer.update(xlayer)
    ydeps = {"ylayer": {"xlayer"}}
    ydeps.update(xdeps)
    hlgy = HighLevelGraph(ylayer, dependencies=ydeps)
    zlayer = {"zlayer": MaterializedLayer({"z": Task("z", func, TaskRef("x"))})}
    zlayer.update(xlayer)
    zdeps = {"zlayer": {"xlayer"}}

    zdeps.update(xdeps)
    hlgz = HighLevelGraph(zlayer, dependencies=zdeps)
    hlgexpry = HLGExpr(
        hlgy,
        output_keys=[["y"], ["x"]],
    )
    hlgexprz = HLGExpr(
        hlgz,
        output_keys=[["z"]],
    )
    expr = _ExprSequence(hlgexprz, hlgexpry)

    expected = {"foo": {"x": 1}}
    assert hlgexprz.__dask_annotations__() == expected
    assert hlgexprz.optimize().__dask_annotations__() == expected
    assert expr.__dask_annotations__() == expected
    assert expr.optimize().__dask_annotations__() == expected

    def only_optimize_with_z(dsk, keys):
        if "z" not in list(flatten(keys)):
            raise RuntimeError("Don't optimize me")
        return optimizer_with_annotations(dsk, keys)

    hlgexpry = HLGExpr(
        hlgy,
        low_level_optimizer=only_optimize_with_z,
        output_keys=[["y"], ["x"]],
    )
    hlgexprz = HLGExpr(
        hlgz,
        low_level_optimizer=only_optimize_with_z,
        output_keys=[["z"]],
    )
    expr = _ExprSequence(hlgexprz, hlgexpry)

    expected = {"foo": {"z-optimized": 1}}
    assert hlgexprz.__dask_annotations__() == expected
    assert hlgexprz.optimize().__dask_annotations__() == expected
    expected = {"foo": {f"{key}-optimized": 1 for key in list("xyz")}}
    with pytest.raises(RuntimeError, match="Don't optimize me"):
        expr.__dask_annotations__()
    assert expr.optimize().__dask_annotations__() == expected


def test_hlg_blockwise_fusion():
    b1 = blockwise(
        lambda x: x,
        "out1",
        "i",
        "x",
        "i",
        numblocks={"x": (2,)},
    )
    b2 = blockwise(
        lambda x: x,
        "out2",
        "i",
        "out1",
        "i",
        numblocks={"out1": (2,)},
    )

    layers = {"b1": b1, "b2": b2}
    deps = {
        "b1": set(),
        "b2": {"b1"},
    }
    hlg = HighLevelGraph(layers, deps)
    # Sanity check that the graph is correct
    assert len(dict(hlg)) == 4
    hlgexpr = HLGExpr(hlg, low_level_optimizer=optimize_blockwise)
    assert len(hlgexpr.__dask_graph__()) == 2
    assert len(hlgexpr.optimize().__dask_graph__()) == 2
    assert len(hlgexpr.finalize_compute().optimize().__dask_graph__()) == 2
